APP_NAME = "intra-search"
STATIC_FOLDER = "../ui/dist"
DEFAULT_PORT = 5000
DEFAULT_CHUNK_SIZE = 50
DEFAULT_MODEL = "msmarco-distilbert-cos-v5"
HOST = "localhost"
SHORT_DESC = "A tool for performing semantic search within documents using sentence transformers to find contextually relevant text."



import os

from flask import Flask, request, send_file, jsonify, send_from_directory
from flask_cors import CORS

from .store import Store
from .model import Model
from .config import STATIC_FOLDER

app = Flask(__name__, static_folder=STATIC_FOLDER)
cors = CORS(app)
store = Store()
_cache = {}


@app.route("/")
def index():
    """Route to serve index.html (entry point)"""
    return send_from_directory(app.static_folder, "index.html")


@app.route("/<path:path>")
def static_proxy(path):
    """Route to serve static assets"""
    return send_from_directory(app.static_folder, path)


@app.route("/api/embeddings")
def get_embeddings_info():
    """GET - all embeddings info from manifest.json"""
    try:
        return store.read_manifest()
    except Exception as e:
        return jsonify({"error": str(e)}), 500


@app.route("/api/doc/<embedding_id>")
def send_doc(embedding_id):
    """GET - serve original document"""
    try:
        embedding_meta = store.get_meta(embedding_id)

        if embedding_meta is None:
            return "Unable to find document embeddings", 404

        document_path = embedding_meta["document_path"]

        return send_file(
            document_path,
            mimetype="application/pdf",
        )
    except FileNotFoundError:
        return "Document not found", 404
    except Exception as e:
        return jsonify({"error": str(e)}), 500


@app.route("/api/<embedding_id>/query")
def query_doc(embedding_id):
    """GET - query document embedding"""
    try:
        embedding = None
        query_string = request.args["query"]

        if embedding_id in _cache:
            embedding = _cache[embedding_id]
        else:
            embedding_meta = store.get_meta(embedding_id)
            if embedding_meta is None:
                return "Unable to find document embeddings", 404

            embedding = store.load(
                os.path.join(store.dir_path, embedding_meta["embedding_name"])
            )

            # Cache the loaded embeddings in-memory
            _cache[embedding_id] = embedding

        model = Model(
            model_name=embedding["model"],
        )

        return model.query(query=query_string, embeddings=embedding["embeddings"])

    except Exception as e:
        return jsonify({"error": str(e)}), 500



import pickle
import os
import json

import platformdirs
import click

from .config import APP_NAME

"""
This module provides the `Store` class for managing document embeddings and metadata.

It handles saving, loading, and deleting embeddings with pickle, while maintaining 
a JSON manifest to track associated metadata.
"""


class Store:

    # os-specific path for caching embeddings
    dir_path = platformdirs.user_data_dir(APP_NAME)
    manifest_path = os.path.join(dir_path, "manifest.json")

    def __init__(self):

        # create directory to store document embeddings
        if not os.path.exists(self.dir_path):
            os.makedirs(self.dir_path)

        # create manifest file
        if not os.path.exists(self.manifest_path):
            with open(self.manifest_path, "w+") as f:
                json.dump([], f)

    def read_manifest(self):
        with open(self.manifest_path, "r") as f:
            return json.load(f)

    def get_meta(self, id):
        meta = list(filter(lambda x: x["id"] == id, self.read_manifest()))
        if len(meta) == 0:
            return None
        return meta[0]

    def _append_manifest(self, item):
        data = self.read_manifest()
        data.append(item)
        with open(self.manifest_path, "w") as f:
            json.dump(data, f, indent=4)

    def save(self, item, meta, file_name):
        pickle.dump(item, open(os.path.join(self.dir_path, file_name), "wb"))
        self._append_manifest(meta)

    def load(self, file_path):
        return pickle.load(open(file_path, "rb"))

    def exist(self, file_path, model_name, chunk_size):
        filter_func = lambda x: (
            x["document_path"] == file_path
            and x["chunk_size"] == chunk_size
            and x["model"] == model_name
        )
        return len(list(filter(filter_func, self.read_manifest()))) > 0

    def delete(self, files):
        meta = self.read_manifest()
        for file in files:

            file_path = os.path.abspath(file)
            file_name = os.path.basename(file)

            occurences = [x for x in meta if x["document_path"] == file_path]
            meta = [x for x in meta if x["document_path"] != file_path]

            if len(occurences) < 1:
                click.secho(f"No embeddings exists for {file_name} ({file_path}).")
                continue

            for ele in occurences:
                embedding_path = os.path.join(self.dir_path, ele["embedding_name"])
                if os.path.isfile(embedding_path):
                    os.remove(embedding_path)

            click.secho(
                f"Deleted all embeddings of {file_name}.",
                fg="green",
            )

        with open(self.manifest_path, "w") as f:
            json.dump(meta, f, indent=4)



"""
This module defines the following CLI commands:
- `create`: Generate document embeddings using specified model and chunk size.
- `start`: Launch the Flask server to serve the API and web application.
- `remove`: Delete existing embeddings based on specified document files.
- `list`: Display a list of all cached embeddings.
"""

import os
from uuid import uuid4

import click

from .server import app
from .doc import Pdf
from .model import Model
from .store import Store
from .utils import sanitize_filename
from .config import DEFAULT_MODEL, DEFAULT_CHUNK_SIZE, DEFAULT_PORT, HOST, SHORT_DESC


@click.group(invoke_without_command=True, help=SHORT_DESC)
@click.pass_context
@click.option(
    "--show-dir",
    "-d",
    "show_dir",
    is_flag=True,
    help="Show the directory where document embeddings are stored.",
)
def cli(ctx, show_dir):
    if ctx.invoked_subcommand is None:
        if show_dir:
            print(Store.dir_path)
            ctx.exit()
    else:
        pass


@click.command()
@click.argument(
    "files",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        path_type=str,
    ),
    nargs=-1,
    required=True,
)
@click.option(
    "--model",
    "-m",
    "model_name",
    default=DEFAULT_MODEL,
    show_default=True,
    help="Embedding model",
)
@click.option(
    "--chunks",
    "-n",
    "chunk_size",
    default=DEFAULT_CHUNK_SIZE,
    show_default=True,
    type=click.IntRange(10, None),
    help="Chunk size, must be greater than 10",
)
def create(files, model_name, chunk_size):
    """Create document embeddings"""
    try:
        completed = []
        store = Store()
        for file in files:

            doc = Pdf(file)

            """ Skip file if embeddings already exists """
            if store.exist(
                file_path=doc.file_path,
                model_name=model_name,
                chunk_size=chunk_size,
            ):
                click.secho(
                    f"Embedding already exists for document - {doc.file_name}",
                    fg="green",
                )
                continue

            model = Model(model_name=model_name)

            click.secho(f"Processing document: {doc.file_name}", bg="blue", fg="white")

            embedding_file_name = (
                sanitize_filename(
                    "_".join(map(str, [doc.file_name, model_name, chunk_size]))
                )
                + ".pkl"
            )

            emeddings_meta = {
                "id": str(uuid4()),
                "model": model_name,
                "chunk_size": chunk_size,
                "document_path": doc.file_path,
                "document_name": doc.file_name,
                "embedding_name": embedding_file_name,
            }

            store.save(
                file_name=embedding_file_name,
                meta=emeddings_meta,
                item={
                    **emeddings_meta,
                    "embeddings": model.get_embeddings(
                        doc.extract_text(chunk_size=chunk_size)
                    ),
                },
            )

            completed.append(doc.file_name)

        cmpl_message = "Embeddings created for the documents:\n"
        cmpl_message += "\n".join(completed)
        click.secho(
            cmpl_message,
            fg="green",
        )
        click.secho(
            "\nRun 'intra-search start' to start the server.\n",
            fg="green",
            bold=True,
        )

    except Exception as e:
        click.secho(f"{e}", err=True)


@click.command()
@click.option(
    "--port",
    "-p",
    default=DEFAULT_PORT,
    show_default=True,
    type=click.INT,
    help="Port of the Flask app (By default, the Flask app runs on port=5000)",
)
def start(port):
    """Start the flask application (which serves both API and web app)"""
    try:
        app.run(host=HOST, port=port)
    except Exception as e:
        click.echo(f"{e}", err=True)
        click.echo(
            f"Try launching the server on a different port by passing --port/-p PORT ",
            bold=True,
        )


@click.command()
@click.argument(
    "files",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        path_type=str,
    ),
    nargs=-1,
    required=True,
)
def remove(files):
    """Remove embeddings"""
    try:
        store = Store()
        cnf_message = "\nThis action will delete all embeddings created using the following documents:\n\n"
        cnf_message += "\n".join(
            [f"{os.path.basename(file)} ({os.path.abspath(file)})" for file in files]
        )
        click.secho(
            cnf_message,
            fg="red",
            bold=True,
        )

        if click.confirm("\n\n Proceed?", default=False, abort=True):
            store.delete(files)

    except Exception as e:
        click.echo(f"{e}", err=True)


@click.command()
def list():
    """List all cached embeddings"""
    try:
        from tabulate import tabulate

        store = Store()
        meta = store.read_manifest()
        header = ["document", "model", "chunk size"]
        rows = [
            [
                ele["document_name"],
                ele["model"],
                ele["chunk_size"],
            ]
            for ele in meta
        ]
        click.echo(tabulate(rows, header))

    except Exception as e:
        click.echo(f"{e}", err=True)


cli.add_command(create)
cli.add_command(start)
cli.add_command(remove)
cli.add_command(list)

if __name__ == "__main__":
    cli()



import re


def get_chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def sanitize(text):
    """
    1. Removes URLs from text
    """
    return re.sub(r"http\S+", "", text)


def sanitize_filename(input_string, char="_"):
    """
    Replace invalid characters with char (default="_")

    """
    invalid_chars_pattern = r'[<>:"/\\|?*.\']'
    sanitized_string = re.sub(invalid_chars_pattern, char, input_string)
    sanitized_string = sanitized_string.strip(char)
    return sanitized_string



import os
from functools import reduce
from uuid import uuid4

import fitz
from tqdm import tqdm

from .utils import get_chunks, sanitize


class Pdf:

    def __init__(self, file_path):

        try:
            self._doc = fitz.open(file_path)
        except:
            raise ("Unable to read file as pdf")

        self.file_path = os.path.abspath(file_path)
        self.file_name = os.path.basename(file_path)

    def _get_words(self):
        """
        Extract and return words from pdf along with page number, page dimensions,
        block number, line number and word sequence number
        """
        words = []
        for page_number, page in enumerate(self._doc):
            words_in_page = [
                (*ele, page_number, page.rect.height, page.rect.width)
                for ele in page.get_text("words")
            ]
            words += words_in_page

        return words

    def _get_boundingrect(self, bboxes):
        """
        Calculate and return the the smallest rectangle that encompasses
        all the bounding boxes for a text passage. If the text spans multiple
        pages, only the bounding boxes from the first page are considered.
        """
        page_number = bboxes[0]["pageNumber"]
        i = 1
        while i < len(bboxes) and page_number == bboxes[i]["pageNumber"]:
            i += 1

        return {
            "x1": bboxes[0]["x1"],
            "y1": bboxes[0]["y1"],
            "x2": bboxes[i - 1]["x2"],
            "y2": bboxes[i - 1]["y2"],
            "pageNumber": page_number,
            "height": bboxes[0]["height"],
            "width": bboxes[0]["width"],
        }

    def _get_text_pos(self, lst):
        """
        Return text positions after grouping word bounding boxes
        based-on page, text block, and line number information.

        lst[i] : tuple(x1, y1, x2, y2, "word", block_no, line_no, word_no, page_number, height, width)
        """
        bboxes = []
        i = 0

        while i < len(lst):

            x1, y1 = lst[i][:2]
            block_no, line_no = lst[i][5:7]
            page_number, height, width = lst[i][8:]

            i += 1

            while (
                i < len(lst)
                and block_no == lst[i][5]
                and line_no == lst[i][6]
                and page_number == lst[i][8]
            ):
                i += 1

            bboxes.append(
                {
                    "x1": x1,
                    "y1": y1,
                    "x2": lst[i - 1][2],
                    "y2": lst[i - 1][3],
                    "pageNumber": page_number + 1,  # convert to 1-indexed
                    "height": height,
                    "width": width,
                }
            )

        return {
            "rects": bboxes,
            "boundingRect": self._get_boundingrect(bboxes),
        }

    def extract_text(self, chunk_size=50):
        embedding_input = []
        words = self._get_words()

        chunks = get_chunks(words, chunk_size)

        for chunk in tqdm(chunks, desc="Extracting text"):
            text = reduce(lambda prev, curr: prev + curr[4] + " ", chunk, "")
            position = self._get_text_pos(chunk)
            embedding_input.append(
                {"id": str(uuid4()), "text": sanitize(text), "position": position}
            )

        return embedding_input



from setuptools import setup, find_packages
from intra_search.config import SHORT_DESC


def read_requirements(file):
    with open(file) as f:
        return f.read().splitlines()


def read_file(file):
    with open(file) as f:
        return f.read()


requirements = read_requirements("./requirements.txt")
# long_description = read_file("README.md")

setup(
    name="intra-search",
    version="0.1.0",
    author="Monish Prabhu",
    author_email="monish.prabhu.official@gmail.com",
    # url = '',
    description=SHORT_DESC,
    # long_description_content_type = "text/markdown",
    # long_description = long_description,
    license="MIT license",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["intra-search = intra_search.cli:cli"]},
    package_data={
        "intra_search": ["../ui/dist/**/*"],
    },
)



from functools import lru_cache

from sentence_transformers import SentenceTransformer, util
from tqdm import tqdm


class Model:

    def __init__(self, model_name):

        self._model = Model.get_model(model_name)

    @staticmethod
    @lru_cache(maxsize=None)
    def get_model(model_name):
        """
        Load and return a SentenceTransformer model with the given model name.
        Caches the model for future use to avoid reloading.
        """
        try:
            return SentenceTransformer(model_name)
        except Exception as e:
            raise SystemExit(
                f"Error occured while retrieving model - {model_name} : {e}"
            )

    def _similarity(self, sim_type="cosine"):
        if sim_type == "cosine":
            return util.cos_sim
        elif sim_type == "dot":
            return util.dot_score
        elif sim_type == "euclid":
            return util.euclidean_sim

    def get_embeddings(self, input):
        """Calculate and return embeddings"""
        for ele in tqdm(input, desc="Computing text embeddings"):
            embedding = self._model.encode(ele["text"])
            ele["embedding"] = embedding
        return input

    def query(self, query, embeddings, sim_type="cosine"):
        """
        Perform a semantic search over embeddings using the specified similarity metric.

        Encodes the query into an embedding and computes its similarity with each
        embedding in the list. Only embeddings with positive similarity scores are
        returned, sorted by similarity in descending order
        """

        query_embedding = self._model.encode(query)
        answers = []

        similarity = self._similarity(sim_type)

        for embedding in embeddings:
            score = similarity(query_embedding, embedding["embedding"]).item()

            # exclude negative similarities
            if score > 0:
                answers.append(
                    {
                        "content": {"text": embedding["text"]},
                        "type": "text",
                        "position": embedding["position"],
                        "similarity": score,
                        "id": embedding["id"],
                    }
                )

        return sorted(answers, key=lambda d: d["similarity"], reverse=True)
